package org.xbib.interlibrary.db.zfl.postgresql;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Collectors;
import org.xbib.datastructures.interpolation.Interpolator;
import org.xbib.interlibrary.api.library.Library;
import org.xbib.interlibrary.api.library.Role;
import org.xbib.interlibrary.api.order.DeliveryScope;
import org.xbib.interlibrary.api.order.Order;
import org.xbib.interlibrary.api.order.Priority;
import org.xbib.interlibrary.api.order.Status;
import org.xbib.interlibrary.api.order.tracking.DisplayStyle;
import org.xbib.interlibrary.api.route.Route;
import org.xbib.interlibrary.common.route.DefaultRoute;
import org.xbib.interlibrary.db.AbstractOrderManager;
import org.xbib.interlibrary.db.DatabaseModule;
import org.xbib.interlibrary.db.table.OrderTable;
import org.xbib.interlibrary.db.table.SummaryOrderTable;
import org.xbib.interlibrary.db.table.TrackTable;
import org.xbib.jdbc.query.Database;
import org.xbib.jdbc.query.Table;
import static org.xbib.interlibrary.api.library.Role.BOTH;
import static org.xbib.interlibrary.api.library.Role.REQUESTER;
import static org.xbib.interlibrary.api.library.Role.SUPPLIER;

public class ZflOrderManager extends AbstractOrderManager {

    private static final Logger logger = Logger.getLogger(ZflOrderManager.class.getName());
    private final String writeOrderStatusWorkOut;

    private final String writeOrderStatusNotAvailable;

    private final String writeOrderStatusShipped;

    public ZflOrderManager(DatabaseModule manager) throws IOException {
        super(manager);
        this.writeOrderStatusWorkOut = fromInputStream("write-order-status-workout.slnp");
        this.writeOrderStatusNotAvailable = fromInputStream("write-order-status-notavailable.slnp");
        this.writeOrderStatusShipped = fromInputStream("write-order-status-shipped.slnp");
    }

    @Override
    public String createOrder(Order order) {
        // TODO
        throw new UnsupportedOperationException();
    }

    @Override
    public Order getOrder(String key) {
        String where = " where m.zf1bestellid = :key";
        String orderby = " order by e8.time desc, e7.time desc, e1.time desc";
        Table table = databaseModule.getSingleRow(readOrderSingleStatement + where + orderby, Map.of("key", key));
        return getOrder(table, 0);
    }

    @Override
    public void updateOrder(Order order, Route route, String signCOmment) {
        // TODO
        throw new UnsupportedOperationException();
    }

    @Override
    public void updateOrderStatus(Order order, Status status, String comment) throws IOException {
        Map.Entry<String, String> entry = toStatusCode(status);
        if (entry == null) {
            logger.log(Level.WARNING, "status code can not be mapped from " + status);
            return;
        }
        String message = null;
        String info = comment != null && !comment.isEmpty() ? "[" + entry.getValue() + "] " + comment : "[" + entry.getValue() + "]";
        Interpolator<String> interpolator = new Interpolator<>();
        switch (entry.getKey()) {
            case "WorkOut" -> {
                interpolator.when().enclosedBy("${").and("}").handleWith((captured, arg) -> switch (captured) {
                    case "supplierisil" -> order.getIsil(SUPPLIER);
                    case "key" -> order.getOrderNumber();
                    case "status" -> entry.getKey();
                    case "substatus" -> entry.getValue();
                    case "info" -> info;
                    default -> throw new IllegalArgumentException("unknown interpolation for: " + captured);
                });
                message = interpolator.interpolate(writeOrderStatusWorkOut, "");
            }
            case "Shipped" -> {
                interpolator.when().enclosedBy("${").and("}").handleWith((captured, arg) -> switch (captured) {
                    case "supplierisil" -> order.getIsil(SUPPLIER);
                    case "key" -> order.getOrderNumber();
                    case "status" -> entry.getKey();
                    case "substatus" -> entry.getValue();
                    case "info" -> info;
                    default -> throw new IllegalArgumentException("unknown interpolation for: " + captured);
                });
                message = interpolator.interpolate(writeOrderStatusShipped, "");
            }
            case "NotAvailable" -> {
                interpolator.when().enclosedBy("${").and("}").handleWith((captured, arg) -> switch (captured) {
                    case "supplierisil" -> order.getIsil(SUPPLIER);
                    case "key" -> order.getOrderNumber();
                    case "status" -> entry.getKey();
                    case "substatus" -> entry.getValue();
                    case "info" -> info;
                    default -> throw new IllegalArgumentException("unknown interpolation for: " + captured);
                });
                message = interpolator.interpolate(writeOrderStatusNotAvailable, "");
            }
            default ->
                    throw new IllegalArgumentException("not implemented: status = " + entry.getKey() + " substatus = " + entry.getValue());
        }
    }

    @Override
    protected Table getOrderHistoryTable(String key) {
        return databaseModule.getTable(readHistoryStatement + " where m.zf1bestellid = :key order by r.zf6lfd asc",
                        Map.of("key", key));
    }

    @Override
    protected void setTrackTableClauses(TrackTable table) {
        if (table.getWhere() == null) {
            table.setWhere(new ArrayList<>());
        }
        if (table.getParams() == null) {
            table.setParams(new HashMap<>());
        }
        if (table.getIsil() != null) {
            table.getWhere().add("req.zf10isil = :isil");
            table.getParams().put("isil", table.getIsil());
        }
        if (table.getClientId() != null) {
            table.getWhere().add("u.zf2bnr = :requesterid");
            table.getParams().put("requesterid", table.getClientId());
        }
        if (table.getCirculationSet() != null) {
            String s = table.getCirculationSet().stream().map(e -> switch (e) {
                case COPY -> "'C'";
                case LOAN -> "'L'";
                case FORM -> "'F'";
                case SCAN -> "'S'";
                case RESERVATION -> "'V'";
            }).collect(Collectors.joining(","));
            if (!s.isEmpty()) {
                table.getWhere().add("m.zf1bstype2 in (" + s + ")");
            }
        }
        if (table.getStatus() != null) {
            List<String> s = translateOrderStatus(table.getStatus());
            if (!s.isEmpty()) {
                table.getWhere().add("trim(s.zf6vorgang) in (" + String.join(",", s) + ")");
            }
        }
        if (table.getFromLastActivityTime() != null && table.getToLastActivityTime() != null) {
            table.getParams().put("fromlastactivitytime", table.getFromLastActivityTime());
            table.getParams().put("tolastactivitytime", table.getToLastActivityTime());
            table.getWhere().add("s.zf6date between trunc( :fromlastactivitytime ) and trunc( :tolastactivitytime )");
        }
        table.setWhereClause(table.getWhere().isEmpty() ? "" : "where " + String.join(" and ", table.getWhere()));
        table.addSort("ordernumber", false);
    }

    protected Table trackOrder(String isil) {
        return databaseModule.getSingleRow(readOrderSingleStatement + " where req.zf10isil = :isil",
                        Map.of("isil", isil));
    }

    @Override
    protected Table trackOrder(String isil, String transactionid) {
        if (isil == null) {
            throw new IllegalArgumentException("isil must not be null");
        }
        if (transactionid == null) {
            throw new IllegalArgumentException("transactionid must not be null");
        }
        return databaseModule.getSingleRow(readOrderSingleStatement +
                        " where" +
                        " (m.zf1bestellid = :transactionid or m.zf1tan = :transactionid)" +
                        " and req.zf10isil = :isil",
                Map.of("isil", isil, "transactionid", transactionid));
    }

    @Override
    protected Table trackOrder(String isil, String clientid, String transactionid) {
        return databaseModule.getSingleRow(readOrderSingleStatement +
                        " where u.zf2bnr = :clientid" +
                        " and (m.zf1bestellid = :transactionid or m.zf1tan = :transactionid)" +
                        " and req.zf10isil = :isil",
                Map.of("isil", isil, "transactionid", transactionid, "clientid", clientid));
    }

    @Override
    protected Table trackOrder(String key, DisplayStyle displayStyle) {
        String types = displayStyle == DisplayStyle.CLIENT ? "and r.zf6vorgang in ('OrderIn', 'WorkOut', 'Shipped', 'Cancel', 'CancelAFL', 'Sign', 'Return', 'PrintRL', 'Finish')" : "";
        return databaseModule.getTable(readHistoryStatement +
                        " where m.zf1bestellid = :key " + types + " order by r.zf6lfd asc",
                        Map.of("key", key));
    }

    private void parseKoerperschaft(String koerperschaft, DefaultRoute route) {
        if (koerperschaft == null) {
            return;
        }
        int pos = koerperschaft.indexOf("$");
        String source = pos > 0 ? koerperschaft.substring(0, pos) : null;
        String sourceInfo = pos > 0 ? koerperschaft.substring(pos + 1) : null;
        if (sourceInfo == null) {
            return;
        }
        String[] bibList = sourceInfo.split("\\|");
        List<String> isils = new ArrayList<>();
        List<String> sigs = new ArrayList<>();
        for (String string : bibList) {
            pos = string.indexOf(":");
            String sigel = pos > 0 ? string.substring(0, pos) : null;
            if (sigel != null) {
                Library library = databaseModule.getInterlibrary().getConfiguration().getLibrary(sigel);
                if (library != null) {
                    isils.add(library.getISIL());
                }
            }
            String sig = pos > 0 ? string.substring(pos + 1) : null;
            if (sig != null) {
                sigs.add(sig);
            }
        }
        route.setNext(isils);
        route.setNextSignatures(sigs);
    }

    /**
     * Parse "tilde format"
     * <p>
     * ~465#NRW OHNESIGNATUR~61#NRW OHNESIGNATUR~HAG4#NRW OHNESIGNATUR~DÃŒ62#??? OHNESIGNATUR~DM13#NRW OHNESIGNATUR~1393#NRW OHNESIGNATUR~1044#NRW OHNESIGNATUR~386#NRW OHNESIGNATUR~6#NRW OHNESIGNATUR~294#NRW OHNESIGNATUR~5#NRW OHNESIGNATUR~929#NRW OHNESIGNATUR
     *
     * @param source the zf8quelle field from ZFL
     * @return cleaned source
     */
    protected String parseSource(String source) {
        if (source == null) {
            return null;
        }
        for (String segment : source.split("~")) {
            int pos = segment.indexOf('#');
            String region = pos > 0 ? segment.substring(pos + 1) : "";
            if (region.contains("NRW")) {
                return "HBZ";
            }
        }
        return source;
    }

    private void createPrevious(DefaultRoute route) {
        List<String> list = new ArrayList<>();
        for (Route.Entry entry : route.getEntries()) {
            if (entry.isActive()) {
                break;
            }
            list.add(entry.getIsil());
        }
        route.setPrevious(list);
    }

    private void createNext(DefaultRoute route) {
        List<String> list = new ArrayList<>();
        boolean ok = false;
        for (Route.Entry entry : route.getEntries()) {
            if (ok) {
                list.add(entry.getIsil());
            }
            if (entry.isActive()) {
                ok = true;
            }
        }
        route.setNext(list);
    }

    protected Table getOrderTable(Database db, String key) {
        return db.getSingleRow(readOrderSingleStatement + " where m.zf1bestellid = :key", Map.of("key", key));
    }

    protected Table getOrderTable(Database db, String key, String isil) {
        return db.getSingleRow(readOrderSingleStatement +
                        " where m.zf1bestellid = :key and (req.zf10isil = :isil or sup.zf10isil = :isil)",
                Map.of("key", key, "isil", isil));
    }

    @Override
    protected void setOrderTableClauses(OrderTable table) {
        if (table.getWhere() == null) {
            table.setWhere(new ArrayList<>());
        }
        if (table.getParams() == null) {
            table.setParams(new LinkedHashMap<>());
        }
        if (table.getSearchParameters().getOrderSystem() != null) {
            table.getWhere().add("'ZFL' = :ordersystem");
            table.getParams().put("ordersystem", table.getSearchParameters().getOrderSystem().name());
        }
        if (table.getSearchParameters().getCirculation() != null) {
            table.getWhere().add("m.zf1bstype2 = :type");
            table.getParams().put("type", Character.toString(table.getSearchParameters().getCirculation().getZflCode()));
        }
        if (table.getSearchParameters().getPriority() != null) {
            String priority = table.getSearchParameters().getPriority().stream()
                    .filter(p -> p.name().startsWith("LP"))
                    .map(p -> "'" + p.name().substring(2) + "'")
                    .collect(Collectors.joining(","));
            if (!priority.isEmpty()) {
                if (table.getSearchParameters().getPriority().contains(Priority.UNDEFINED)) {
                    table.getWhere().add("(zl.zf7katkeyalpha is null or zl.zf7katkeyalpha in (" + priority + "))");
                } else {
                    table.getWhere().add("zl.zf7katkeyalpha in (" + priority + ")");
                }
            } else {
                if (table.getSearchParameters().getPriority().contains(Priority.UNDEFINED)) {
                    table.getWhere().add("zl.zf7katkeyalpha is null");
                }
            }
        }
        DeliveryScope deliveryScope = table.getSearchParameters().getDeliveryScope();
        if (deliveryScope != null) {
            if (deliveryScope == DeliveryScope.EXTERNAL) {
                table.getWhere().add("sup.zf10isil in ('SWB', 'GBV', 'KOBV', 'HEBIS', 'BVB', 'DE-576', 'DE-601', 'DE-602', 'DE-603', 'DE-604')");
            } else {
                table.getWhere().add("sup.zf10isil not in ('SWB', 'GBV', 'KOBV', 'HEBIS', 'BVB', 'DE-576', 'DE-601', 'DE-602', 'DE-603', 'DE-604')");
            }
        }
        if (table.getSearchParameters().getStatus() != null) {
            List<String> statusnames = new ArrayList<>();
            int statusnamecounter = 1;
            for (String string : translateOrderStatus(table.getSearchParameters().getStatus())) {
                statusnames.add(":status" + statusnamecounter);
                table.getParams().put("status" + statusnamecounter, string);
                statusnamecounter++;
            }
            if (!statusnames.isEmpty()) {
                table.getWhere().add("trim(m.zf1status) in (" + String.join(",", statusnames) + ")");
            }
            List<String> substatusnames = new ArrayList<>();
            int substatusnamecounter = 1;
            for (String string : translateOrderSubStatus(table.getSearchParameters().getStatus())) {
                substatusnames.add(":substatus" + substatusnamecounter);
                table.getParams().put("substatus" + substatusnamecounter, string);
                substatusnamecounter++;
            }
            if (!substatusnames.isEmpty()) {
                table.getWhere().add("trim(s.zf6local) in (" + String.join(",", substatusnames) + ")");
            }
        }
        Role role = table.getSearchParameters().getRole();
        if (role == REQUESTER && table.getSearchParameters().getISIL() != null) {
            table.getWhere().add("req.zf10isil = :isil");
            table.getParams().put("isil", table.getSearchParameters().getISIL());
        } else if (role == SUPPLIER && table.getSearchParameters().getISIL() != null) {
            table.getWhere().add("sup.zf10isil = :isil");
            table.getParams().put("isil", table.getSearchParameters().getISIL());
        } else if (role == BOTH && table.getSearchParameters().getISIL() != null) {
            table.getWhere().add("sup.zf10isil = :isil or req.zf10isil = :isil");
            table.getParams().put("isil", table.getSearchParameters().getISIL());
        }
        if (table.getSearch() != null) {
            if (!table.getSearch().contains("%")) {
                table.getWhere().add("(m.zf1bestellid like :search or m.zf1tan like :search)");
                table.getParams().put("search", table.getSearch());
            } else {
                table.getWhere().add("(m.zf1bestellid = :search or m.zf1tan = :search)");
                table.getParams().put("search", table.getSearch());
            }
        }
        if (table.getFromOrderTime() != null && table.getToOrderTime() != null) {
            table.getParams().put("fromordertime", table.getFromOrderTime());
            table.getParams().put("toordertime", table.getToOrderTime());
            table.getWhere().add("i.zf6date between trunc( :fromordertime ) and trunc( :toordertime )");
        }
        if (table.getFromLastActivityTime() != null && table.getToLastActivityTime() != null) {
            table.getParams().put("fromlastactivitytime", table.getFromLastActivityTime());
            table.getParams().put("tolastactivitytime", table.getToLastActivityTime());
            table.getWhere().add("s.zf6date between trunc( :fromlastactivitytime ) and trunc( :tolastactivitytime )");
        }
        if (table.getFromSupplyTime() != null && table.getToSupplyTime() != null) {
            table.getParams().put("fromsupplytime", table.getFromSupplyTime());
            table.getParams().put("tosupplytime", table.getToSupplyTime());
            table.getWhere().add("s.zf6date between trunc( :fromsupplytime ) and trunc( :tosupplytime )");
        }
        if (table.getSearchParameters().getOffset() != null) {
            table.setOffset(table.getSearchParameters().getOffset());
        }
        if (table.getSearchParameters().getSize() != null) {
            table.setSize(table.getSearchParameters().getSize());
        }
        table.setWhereClause(table.getWhere().isEmpty() ? "" : "where " + String.join(" and ", table.getWhere()));
    }

    @Override
    protected void setSummaryOrderTableClauses(SummaryOrderTable table) {
        if (table.getWhere() == null) {
            table.setWhere(new ArrayList<>());
        }
        if (table.getParams() == null) {
            table.setParams(new HashMap<>());
        }
        if (table.getRole()  == REQUESTER && table.getIsil() != null) {
            table.getWhere().add("req.it_isil = :isil");
            table.getParams().put("isil", table.getIsil());
        } else if (table.getRole() == SUPPLIER && table.getIsil() != null) {
            table.getWhere().add("sup.it_isil = :isil");
            table.getParams().put("isil", table.getIsil());
        }
        if (table.getCirculationSet() != null) {
            String s = table.getCirculationSet().stream().map(e -> switch (e) {
                        case COPY -> "'C'";
                        case LOAN -> "'L'";
                        case FORM -> "'F'";
                        case SCAN -> "'S'";
                        case RESERVATION -> "'V'";
                    })
                    .collect(Collectors.joining(","));
            if (!s.isEmpty()) {
                table.getWhere().add("m.zf1bstype2 in (" + s + ")");
            }
        }
        if (table.getFromOrderTime() != null && table.getToOrderTime() != null) {
            table.getParams().put("fromordertime", table.getFromOrderTime());
            table.getParams().put("toordertime", table.getToOrderTime());
            table.getWhere().add("i.zf6date between trunc( :fromordertime ) and trunc( :toordertime )");
        }
        String where = table.getWhere().isEmpty() ? "" : "where " + String.join(" and ", table.getWhere());
        table.setStatement(String.format(table.getStatement(), where));
        table.setGroupByClause("""
                group by 'FL', decode(m.zf1bstype2,
                      'C', 'COPY',
                      'L', 'LOAN',
                      'F', 'FORM',
                      'V', 'RESERVATION',
                      m.zf1bstype2), trim(s.zf6vorgang), trim(s.zf6local)
""");
        table.addSort("circulation", true);
        table.addSort("status", true);
        table.addSort("substatus", true);
    }

    private static List<String> translateOrderStatus(Status status) {
        List<String> list = new ArrayList<>();
        switch (status.getCode()) {
            case INCOMING -> list.add("OrderIn");
            case PROCESSING -> list.add("WorkOut");
            case SHIP -> list.add("Shipped");
            case NOTSUPPLIED -> list.add("NotAvailable");
            case CANCELLED -> {
                list.add("Cancel");
                list.add("CancelAFL");
            }
            case SIGN -> list.add("Sign");
            case PROCESSED -> list.add("PrintRL");
            case RETURN -> list.add("Return");
            case EXPIRE -> list.add("Expire");
            case END -> list.add("Finish");
            case QUIT -> list.add("Quit");
            case ERROR -> list.add("Failure");
            // ROUTING
        }
        return list;
    }

    private static List<String> translateOrderSubStatus(Status status) {
        List<String> list = new ArrayList<>();
        switch (status.getSubstatus()) {
            case INFO -> list.add("info");
            case INPROCESS -> list.add("inprocess");
            case READY -> list.add("ready");
            case PROBED -> list.add("probed");
            case NOTIFIED -> list.add("notified");
            case ERROR -> list.add("error");
            case COMMENT -> list.add("comment");
            case REQUESTFORM -> list.add("requestform");
            case TRANSFERRED -> list.add("transferred");
            case NOTTRANSFERRED -> list.add("nottransferred");
        }
        return list;
    }

    private Map.Entry<String, String> toStatusCode(Status status) {
        Map.Entry<String, String> entry = null;
        switch (status.getCode()) {
            case INCOMING -> {
                entry = Map.entry("OrderIn", "");
            }
            case PROCESSING -> {
                switch (status.getSubstatus()) {
                    case REQUESTFORM -> entry = Map.entry("WorkOut", "requestform");
                    case TRANSFERRED -> entry = Map.entry("WorkOut", "transferred");
                    case NOTTRANSFERRED -> entry = Map.entry("WorkOut", "nottransferred");
                    default -> entry = Map.entry("WorkOut", "");
                }
            }
            case SHIP -> {
                switch (status.getSubstatus()) {
                    case INPROCESS -> entry = Map.entry("Shipped", "inprocess");
                    case READY -> entry = Map.entry("Shipped", "ready");
                    case PROBED -> entry = Map.entry("Shipped", "probed");
                    case NOTIFIED -> entry = Map.entry("Shipped", "notified");
                    default -> entry = Map.entry("Shipped", "");
                }
            }
            case NOTSUPPLIED -> entry = Map.entry("NotAvailable", "");
            case CANCELLED -> entry = Map.entry("CancelAFL", "");
            case ERROR -> entry = Map.entry("Shipped", "error");
            case SIGN -> entry = Map.entry("Sign", "");
            // unsupported: UNDEFINED, PREPARING, ROUTING, RETURN, END, EXPIRE, QUIT, INFO
        }
        return entry;
    }
}
