package org.xbib.interlibrary.db.zfl.postgresql;

import java.io.IOException;
import java.math.BigDecimal;
import java.nio.charset.StandardCharsets;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.xbib.datastructures.json.tiny.Json;
import org.xbib.interlibrary.api.library.Library;
import org.xbib.interlibrary.api.library.LibraryService;
import org.xbib.interlibrary.api.order.Circulation;
import org.xbib.interlibrary.api.route.Route;
import org.xbib.interlibrary.common.route.DefaultRoute;
import org.xbib.interlibrary.http.HttpHelper;
import org.xbib.interlibrary.db.AbstractRouteManager;
import org.xbib.interlibrary.db.DatabaseModule;
import org.xbib.jdbc.query.Table;

public class ZflRouteManager extends AbstractRouteManager {

    private static final Logger logger = Logger.getLogger(ZflRouteManager.class.getName());

    private final HttpHelper httpClient;

    private final String availUrl;

    public ZflRouteManager(DatabaseModule manager,
                           HttpHelper httpClient) throws IOException {
        super(manager);
        this.httpClient = Objects.requireNonNull(httpClient);
        this.availUrl = Objects.requireNonNull(databaseModule.getInterlibrary().getSettings()
                .get("database.url.avail"), "database.url.avail not defined in settings");
    }

    @SuppressWarnings("unchecked")
    @Override
    protected Route getRoute(Table table) {
        if (table.isEmpty()) {
            return null;
        }
        DefaultRoute route = new DefaultRoute();
        String requesterisil = table.getValue(0, "requesterisil");
        String source = table.getValue(0, "source");
        source = databaseModule.parseSource(source);
        String sourceid = table.getValue(0, "sourceid");
        String yearString = table.getValue(0, "year");
        Map<String, Object> availResponse = null;
        try {
            availResponse = avail(availUrl, Map.of("requester", requesterisil,
                    "source", source,
                    "sourceid", sourceid,
                    "year", yearString,
                    "self_allowed", "true",
                    "local_allowed", "true",
                    "domain_allowed", "true")
            );
        } catch (IOException e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
        }
        // always skip i = 0, this position ist for "PFL-Verbuchung"
        for (int i = 1; i < table.getRowCount(); i++) {
            BigDecimal pos = table.getValue(i, "pos");
            String active = table.getValue(i, "active");
            LocalDateTime time = table.getValue(i, "time");
            String transportservice = table.getValue(i, "transportservice");
            Circulation circulation = Circulation.valueOf(table.getValue(i, "circulation"));
            String supplierisil = table.getValue(i, "supplierisil");
            Library library = databaseModule.getInterlibrary().getConfiguration().getLibrary(supplierisil);
            if (library != null) {
                if (!supplierisil.startsWith("DE")) {
                    supplierisil = library.getISIL();
                }
                String name = library.getName();
                String domain = library.getDomain().getName();
                String organization = library.getDomain().getOrganization();
                boolean isOrganization = library.isOrganization();
                List<LibraryService> services = new ArrayList<>();
                if (availResponse != null) {
                    Map<String, Object> meta = (Map<String, Object>) availResponse.get("meta");
                    if (meta != null) {
                        if (meta.containsKey("link")) {
                            route.setLinks((List<Map<String, Object>>) meta.get("link"));
                        }
                        if (meta.containsKey("linkedindicators")) {
                            route.setLinkedIndicators((List<Map<String, Object>>) meta.get("linkedindicators"));
                        }
                    }
                    services.addAll(createServices(supplierisil, (Map<String, Object>) availResponse.get("interlibrary")));
                    services.addAll(createServices(supplierisil, (Map<String, Object>) availResponse.get("interlibrarydomains")));
                }
                route.addEntry(pos.intValue(), "Y".equals(active), time, supplierisil, name, domain, organization,
                        isOrganization, transportservice, circulation, services);
            } else {
                logger.log(Level.WARNING, "no library found for " + supplierisil);
            }
        }
        // one source of next/nextSignatures is field "Koerperschaft" but that is a hack so we skip it
        //parseKoerperschaft(table.getValue(0, "koerperschaft"), route);
        createPrevious(route);
        createNext(route);
        return route;
    }

    private void createPrevious(DefaultRoute route) {
        List<String> list = new ArrayList<>();
        for (Route.Entry entry : route.getEntries()) {
            if (entry.isActive()) {
                break;
            }
            list.add(entry.getIsil());
        }
        route.setPrevious(list);
    }

    private void createNext(DefaultRoute route) {
        List<String> list = new ArrayList<>();
        boolean ok = false;
        for (Route.Entry entry : route.getEntries()) {
            if (ok) {
                list.add(entry.getIsil());
            }
            if (entry.isActive()) {
                ok = true;
            }
        }
        route.setNext(list);
    }

    private Map<String, Object> avail(String availUrl, Map<String, Object> map) throws IOException {
        String json = availAsJson(availUrl, map);
        return json.isEmpty() ? Map.of() : Json.toMap(json);
    }

    private String availAsJson(String url, Map<String, Object> map) throws IOException {
            StringBuilder sb = new StringBuilder();
            httpClient.get(url,
                    params -> {
                        if (map != null) {
                            params.putAll(map);
                        }
                    },
                    headers -> {
                        headers.add("accept","application/json");
                    },
                    response -> {
                        int status = response.getStatus().code();
                        if (status == 200) {
                            sb.append(response.getBodyAsChars(StandardCharsets.UTF_8).toString());
                        } else {
                            logger.log(Level.SEVERE, "HTTP response status = " + status);
                        }
                    },
                    exception -> logger.log(Level.SEVERE, exception.getMessage(), exception),
                    timeout -> logger.log(Level.WARNING, "timeout while waiting for response for " + url));
            return sb.toString();
        }
}
